import numpy as np
from scipy.special import kn
import properties
from ....utils.code_utils import deprecate_class

from ....utils import mkvc, sdiag, Zero
from ...base import BaseEMSimulation
from ....data import Data

from .survey import Survey
from . import sources
from . import receivers
from .fields_2d import Fields2D, Fields2DCellCentered, Fields2DNodal
from .fields import FieldsDC, Fields3DCellCentered, Fields3DNodal
from .boundary_utils import getxBCyBC_CC


class BaseDCSimulation2D(BaseEMSimulation):
    """
    Base 2.5D DC problem
    """

    survey = properties.Instance(
        "a DC survey object", Survey, required=True
    )

    storeJ = properties.Bool(
        "store the sensitivity", default=False
    )

    fieldsPair = Fields2D  # SimPEG.EM.Static.Fields_2D
    fieldsPair_fwd = FieldsDC
    nky = 15
    kys = np.logspace(-4, 1, nky)
    Ainv = [None for i in range(nky)]
    nT = nky-1  # Only for using TimeFields
    # there's actually nT+1 fields, so we don't need to store the last one
    _Jmatrix = None
    fix_Jmatrix = False
    _mini_survey = None

    def __init__(self, *args, **kwargs):
        miniaturize = kwargs.pop('miniaturize', False)
        super().__init__(*args, **kwargs)
        # Do stuff to simplify the forward and JTvec operation if number of dipole
        # sources is greater than the number of unique pole sources
        if miniaturize:
            survey = self.survey
            survey.getABMN_locations()
            A = survey.a_locations
            B = survey.b_locations
            M = survey.m_locations
            N = survey.n_locations
            u_src, inv_src = np.unique(np.r_[A, B], axis=0, return_inverse=True)
            if len(u_src) < survey.nSrc:
                if self.verbose:
                    print('Number of sources greater than number of source electrodes.')
                    print('Switching internally to smaller survey.')

                inv_A, inv_B = inv_src.reshape(2, -1)
                ind_dipoles_tx = inv_A != inv_B

                unique_sources = []
                #ranges = np.arange(survey.nD*2)
                pos_inds = np.zeros(survey.nD, dtype=int)
                neg_inds = np.zeros(survey.nD, dtype=int)
                counter = 0
                for i_src, src_loc in enumerate(u_src):
                    # add all the receivers which used that pole.
                    # for each dipole source, essentialy it creates 2 receivers for that source.
                    # Even though there's "double" the number of receivers, the number of sources
                    # is reduced to the minimum necessary to reconstruct the survey. Reducing the memory
                    # required to store the fields, and allowing more rhs solutions to be done in parallel
                    # when storing the J matrix.

                    ind_A = np.where(inv_A == i_src)[0]  # these are all individually unique
                    ind_B = np.where(inv_B == i_src)[0]  #    "

                    # There will be overlapping indices in ind_A and ind_B if there are pole
                    # sources in the original survey.
                    ind_AB, _is = np.unique(np.r_[ind_A, ind_B], return_inverse=True)

                    ind_AB_to_A = _is[:len(ind_A)]
                    ind_AB_to_B = _is[len(ind_A):]

                    my_ms = M[ind_AB]
                    my_ns = N[ind_AB]
                    rx_dipole_inds = np.any(my_ms != my_ns, axis=1)

                    recs = []
                    my_ns = my_ns[rx_dipole_inds]
                    n_dipole_rx = len(my_ns)
                    if n_dipole_rx > 0:
                        recs.append(receivers.Dipole2D(my_ms[rx_dipole_inds], my_ns))

                    my_ms = my_ms[~rx_dipole_inds]
                    n_pole_rx = len(my_ms)
                    if n_pole_rx > 0:
                        recs.append(receivers.Pole2D(my_ms))

                    my_data_inds = np.arange(len(ind_AB)) + counter
                    counter += len(ind_AB)

                    my_plus_inds = my_data_inds[ind_AB_to_A]
                    my_neg_inds = my_data_inds[ind_AB_to_B]

                    pos_inds[ind_A] = my_plus_inds
                    neg_inds[ind_B] = my_neg_inds

                    unique_sources.append(sources.Pole(recs, src_loc))
                # only keep the negative indices of tx's that had a dipole source
                self._pos_inds = pos_inds
                self._neg_inds = neg_inds[ind_dipoles_tx]
                self._ind_dipole_txs = ind_dipoles_tx
                self._mini_survey = Survey(unique_sources)

    def set_geometric_factor(self, geometric_factor):
        index = 0
        for src in self.survey.source_list:
            for rx in src.receiver_list:
                rx._geometric_factor = geometric_factor[index]
                index += 1

    def fields(self, m):
        if self.verbose:
            print (">> Compute fields")
        if m is not None:
            self.model = m
        if self.Ainv[0] is not None:
            for i in range(self.nky):
                self.Ainv[i].clean()
        f = self.fieldsPair(self)
        for iky, ky in enumerate(self.kys):
            A = self.getA(ky)
            self.Ainv[iky] = self.Solver(A, **self.solver_opts)
            RHS = self.getRHS(ky)
            u = self.Ainv[iky] * RHS
            f[:, self._solutionType, iky] = u
        return f

    def fields_to_space(self, f, y=0.):
        f_fwd = self.fieldsPair_fwd(self.mesh, self.survey)
        # Evaluating Integration using Trapezoidal rules
        dky = np.diff(self.kys)/2
        trap_weights = np.r_[dky, 0]+np.r_[0, dky]
        trap_weights *= np.cos(self.kys*y)
        # assume constant value at 0 frequency?
        trap_weights[0] += self.kys[0]/2 * (1.0 + np.cos(self.kys[0]*y))
        trap_weights /= np.pi
        phi = f[:, self._solutionType, :].dot(trap_weights)
        f_fwd[:, self._solutionType] = phi
        return f_fwd

    def dpred(self, m=None, f=None):
        """
        Project fields to receiver locations
        :param Fields u: fields object
        :rtype: numpy.ndarray
        :return: data
        """
        if f is None:
            if m is None:
                m = self.model
            f = self.fields(m)

        kys = self.kys
        if self._mini_survey is not None:
            survey = self._mini_survey
        else:
            survey = self.survey

        temp = np.empty(survey.nD)
        count = 0
        for src in survey.source_list:
            for rx in src.receiver_list:
                d = rx.eval(kys, src, self.mesh, f)
                temp[count:count+len(d)] = d
                count += len(d)

        return self._mini_survey_data(temp)

    def getJ(self, m, f=None):
        """
            Generate Full sensitivity matrix
        """
        if self._Jmatrix is not None:
            return self._Jmatrix
        else:
            if self.verbose:
                print("Calculating J and storing")
            self.model = m
            if f is None:
                f = self.fields(m)
            self._Jmatrix = (self._Jtvec(m, v=None, f=f)).T
        return self._Jmatrix

    def Jvec(self, m, v, f=None):
        """
            Compute sensitivity matrix (J) and vector (v) product.
        """
        if self.storeJ:
            J = self.getJ(m, f=f)
            Jv = mkvc(np.dot(J, v))
            return Jv

        self.model = m

        if f is None:
            f = self.fields(m)

        # TODO: This is not a good idea !! should change that as a list
        if self._mini_survey is not None:
            survey = self._mini_survey
        else:
            survey = self.survey

        Jv = np.zeros(survey.nD)
        # Assume y=0.
        # This needs some thoughts to implement in general when src is dipole
        y = 0.
        dky = np.diff(self.kys)/2
        trap_weights = np.r_[dky, 0]+np.r_[0, dky]
        trap_weights *= np.cos(self.kys*y)  # *(1.0/np.pi)
        # assume constant value at 0 frequency?
        trap_weights[0] += self.kys[0]/2 * (1.0 + np.cos(self.kys[0]*y))
        trap_weights /= np.pi

        # TODO: this loop is pretty slow .. (Parellize)
        for iky, ky in enumerate(self.kys):
            u_ky = f[:, self._solutionType, iky]
            count = 0
            for i_src, src in enumerate(survey.source_list):
                u_src = u_ky[:, i_src]
                dA_dm_v = self.getADeriv(ky, u_src, v, adjoint=False)
                # dRHS_dm_v = self.getRHSDeriv(ky, src, v) = 0
                du_dm_v = self.Ainv[iky] * (-dA_dm_v)  # + dRHS_dm_v)
                for rx in src.receiver_list:
                    df_dmFun = getattr(f, '_{0!s}Deriv'.format(rx.projField),
                                       None)
                    df_dm_v = df_dmFun(iky, src, du_dm_v, v, adjoint=False)
                    Jv1_temp = rx.evalDeriv(ky, src, self.mesh, f, df_dm_v)
                    # Trapezoidal intergration
                    Jv[count:count+len(Jv1_temp)] += trap_weights[iky]*Jv1_temp
                    count += len(Jv1_temp)

        return self._mini_survey_data(Jv)

    def Jtvec(self, m, v, f=None):
        """
            Compute adjoint sensitivity matrix (J^T) and vector (v) product.
        """
        if self.storeJ:
            J = self.getJ(m, f=f)
            Jtv = mkvc(np.dot(J.T, v))
            return Jtv

        self.model = m

        if f is None:
            f = self.fields(m)

        return self._Jtvec(m, v=v, f=f)

    def _Jtvec(self, m, v=None, f=None):
        """
            Compute adjoint sensitivity matrix (J^T) and vector (v) product.
            Full J matrix can be computed by inputing v=None
        """

        # Assume y=0.
        # This needs some thoughts to implement in general when src is dipole
        y = 0.
        dky = np.diff(self.kys)/2
        trap_weights = np.r_[dky, 0]+np.r_[0, dky]
        trap_weights *= np.cos(self.kys*y)  # *(1.0/np.pi)
        # assume constant value at 0 frequency?
        trap_weights[0] += self.kys[0]/2 * (1.0 + np.cos(self.kys[0]*y))
        trap_weights /= np.pi
        if self._mini_survey is not None:
            survey = self._mini_survey
        else:
            survey = self.survey

        if v is not None:
            # Ensure v is a data object.
            if isinstance(v, Data):
                v = v.dobs
            v = self._mini_survey_dataT(v)
            Jtv = np.zeros(m.size, dtype=float)

            # TODO: this loop is pretty slow .. (Parellize)
            for iky, ky in enumerate(self.kys):
                u_ky = f[:, self._solutionType, iky]
                count = 0
                for i_src, src in enumerate(survey.source_list):
                    u_src = u_ky[:, i_src]
                    df_duT_sum = 0
                    df_dmT_sum = 0
                    for rx in src.receiver_list:
                        my_v = v[count:count+rx.nD]
                        count += rx.nD
                        # wrt f, need possibility wrt m
                        PTv = rx.evalDeriv(ky, src, self.mesh, f, my_v,
                                           adjoint=True)
                        df_duTFun = getattr(
                            f, '_{0!s}Deriv'.format(rx.projField), None
                        )
                        df_duT, df_dmT = df_duTFun(iky, src, None, PTv,
                                                   adjoint=True)
                        df_duT_sum += df_duT
                        df_dmT_sum += df_dmT

                    ATinvdf_duT = self.Ainv[iky] * df_duT_sum

                    dA_dmT = self.getADeriv(ky, u_src, ATinvdf_duT,
                                            adjoint=True)
                    # dRHS_dmT = self.getRHSDeriv(ky, src, ATinvdf_duT,
                    #                            adjoint=True)
                    du_dmT = -dA_dmT  # + dRHS_dmT=0
                    Jtv += trap_weights[iky]*(df_dmT + du_dmT).astype(float)
            return mkvc(Jtv)

        else:
            # This is for forming full sensitivity matrix
            Jt = np.zeros((self.model.size, survey.nD), order='F')
            for iky, ky in enumerate(self.kys):
                u_ky = f[:, self._solutionType, iky]
                istrt = 0
                for i_src, src in enumerate(survey.source_list):
                    u_src = u_ky[:, i_src]
                    for rx in src.receiver_list:
                        # wrt f, need possibility wrt m
                        P = rx.getP(self.mesh, rx.projGLoc(f)).toarray()

                        ATinvdf_duT = self.Ainv[iky] * (P.T)

                        dA_dmT = self.getADeriv(ky, u_src, ATinvdf_duT,
                                                adjoint=True)
                        Jtv = -trap_weights[iky]*dA_dmT #RHS=0
                        iend = istrt + rx.nD
                        if rx.nD == 1:
                            Jt[:, istrt] += Jtv
                        else:
                            Jt[:, istrt:iend] += Jtv
                        istrt += rx.nD
            return (self._mini_survey_data(Jt.T)).T

    def getSourceTerm(self, ky):
        """
        takes concept of source and turns it into a matrix
        """
        """
        Evaluates the sources, and puts them in matrix form
        :rtype: (numpy.ndarray, numpy.ndarray)
        :return: q (nC or nN, nSrc)
        """

        if self._mini_survey is not None:
            Srcs = self._mini_survey.source_list
        else:
            Srcs = self.survey.source_list

        if self._formulation == 'EB':
            n = self.mesh.nN
            # return NotImplementedError

        elif self._formulation == 'HJ':
            n = self.mesh.nC

        q = np.zeros((n, len(Srcs)), order='F')

        for i, src in enumerate(Srcs):
            q[:, i] = src.eval(self)
        return q

    @property
    def deleteTheseOnModelUpdate(self):
        toDelete = super(BaseDCSimulation2D, self).deleteTheseOnModelUpdate
        if self.sigmaMap is not None:
            toDelete += [
                '_MnSigma', '_MnSigmaDerivMat',
                '_MccRhoi', '_MccRhoiDerivMat'
            ]

        if self.fix_Jmatrix:
            return toDelete

        if self._Jmatrix is not None:
            toDelete += ['_Jmatrix']
        return toDelete

    def _mini_survey_data(self, d_mini):
        if self._mini_survey is not None:
            out = d_mini[self._pos_inds]
            out[self._ind_dipole_txs] -= d_mini[self._neg_inds]
        else:
            out = d_mini
        return out

    def _mini_survey_dataT(self, v):
        if self._mini_survey is not None:
            out = np.zeros(self._mini_survey.nD)
            out[self._pos_inds] = v
            out[self._neg_inds] -= v[self._ind_dipole_txs]
            return out
        else:
            out = v
        return out

    ####################################################
    # Mass Matrices
    ####################################################

    @property
    def MnSigma(self):
        """
            Node inner product matrix for \\(\\sigma\\). Used in the E-B
            formulation
        """
        # TODO: only works isotropic sigma
        if getattr(self, '_MnSigma', None) is None:
            sigma = self.sigma
            vol = self.mesh.vol
            self._MnSigma = sdiag(
                self.mesh.aveN2CC.T*(vol*sigma)
            )
        return self._MnSigma

    @property
    def MnSigmaDerivMat(self):
        """
            Derivative of MnSigma with respect to the model
        """
        if getattr(self, '_MnSigmaDerivMat', None) is None:
            vol = self.mesh.vol
            self._MnSigmaDerivMat = (
                self.mesh.aveN2CC.T * sdiag(vol) * self.sigmaDeriv
                )
        return self._MnSigmaDerivMat

    def MnSigmaDeriv(self, u, v, adjoint=False):
        """
            Derivative of MnSigma with respect to the model times a vector (u)
        """
        if v.ndim > 1:
            u = u[:, None]
        if self.storeInnerProduct:
            if adjoint:
                return self.MnSigmaDerivMat.T * (u * v)
            else:
                return u*(self.MnSigmaDerivMat * v)
        else:
            vol = self.mesh.vol
            if v.ndim > 1:
                vol = vol[:, None]
            if adjoint:
                return self.sigmaDeriv.T * (
                    vol * (self.mesh.aveN2CC * (u * v))
                )
            else:
                dsig_dm_v = self.sigmaDeriv * v
                return (
                    u * (self.mesh.aveN2CC.T * (vol * dsig_dm_v))
                )

    @property
    def MccRhoi(self):
        """
            Cell inner product matrix for \\(\\rho^{-1}\\). Used in the H-J
            formulation
        """
        # TODO: only works isotropic rho
        if getattr(self, '_MccRhoi', None) is None:
            self._MccRhoi = sdiag(
                self.mesh.vol/self.rho
            )
        return self._MccRhoi

    @property
    def MccRhoiDerivMat(self):
        """
            Derivative of MccRho with respect to the model
        """
        if getattr(self, '_MccRhoiDerivMat', None) is None:
            rho = self.rho
            vol = self.mesh.vol
            self._MccRhoiDerivMat = (
                sdiag(vol*(-1./rho**2))*self.rhoDeriv
            )
        return self._MccRhoiDerivMat

    def MccRhoiDeriv(self, u, v, adjoint=False):
        """
            Derivative of :code:`MccRhoi` with respect to the model.
        """
        if self.rhoMap is None:
            return Zero()

        if len(self.rho.shape) > 1:
            if self.rho.shape[1] > self.mesh.dim:
                raise NotImplementedError(
                    "Full anisotropy is not implemented for MccRhoiDeriv."
                )
        if self.storeInnerProduct:
            if adjoint:
                return self.MccRhoiDerivMat.T * (sdiag(u) * v)
            else:
                return sdiag(u) * (self.MccRhoiDerivMat * v)
        else:
            vol = self.mesh.vol
            rho = self.rho
            if adjoint:
                return self.rhoDeriv.T * (sdiag(u*vol*(-1./rho**2)) * v)
            else:
                return (sdiag(u*vol*(-1./rho**2)))*(self.rhoDeriv * v)


class Simulation2DCellCentered(BaseDCSimulation2D):
    """
    2.5D cell centered DC problem
    """

    _solutionType = 'phiSolution'
    _formulation = 'HJ'  # CC potentials means J is on faces
    fieldsPair = Fields2DCellCentered
    fieldsPair_fwd = Fields3DCellCentered
    bc_type = 'Mixed'

    def __init__(self, mesh, **kwargs):
        BaseDCSimulation2D.__init__(self, mesh, **kwargs)

    def getA(self, ky):
        """
        Make the A matrix for the cell centered DC resistivity problem
        A = D MfRhoI G
        """
        # To handle Mixed boundary condition
        self.setBC(ky=ky)
        D = self.Div
        G = self.Grad
        vol = self.mesh.vol
        MfRhoI = self.MfRhoI
        # Get resistivity rho
        rho = self.rho
        A = D * MfRhoI * G + ky**2 * self.MccRhoi
        if self.bc_type == "Neumann":
            A[0, 0] = A[0, 0] + 1.
        return A

    def getADeriv(self, ky, u, v, adjoint=False):
        # To handle Mixed boundary condition
        # self.setBC(ky=ky)

        D = self.Div
        G = self.Grad
        if adjoint:
            return (
                self.MfRhoIDeriv(G*u.flatten(), D.T*v, adjoint=adjoint) +
                ky**2 * self.MccRhoiDeriv(u.flatten(), v, adjoint=adjoint)
            )
        else:
            return (
                D * self.MfRhoIDeriv(G*u.flatten(), v, adjoint=adjoint) +
                ky**2 * self.MccRhoiDeriv(u.flatten(), v, adjoint=adjoint)
            )

    def getRHS(self, ky):
        """
        RHS for the DC problem
        q
        """

        RHS = self.getSourceTerm(ky)
        return RHS

    def getRHSDeriv(self, ky, src, v, adjoint=False):
        """
        Derivative of the right hand side with respect to the model
        """
        # TODO: add qDeriv for RHS depending on m
        # qDeriv = src.evalDeriv(self, ky, adjoint=adjoint)
        # return qDeriv
        return Zero()

    def setBC(self, ky=None):
        fxm, fxp, fym, fyp = self.mesh.faceBoundaryInd
        gBFxm = self.mesh.gridFx[fxm, :]
        gBFxp = self.mesh.gridFx[fxp, :]
        gBFym = self.mesh.gridFy[fym, :]
        gBFyp = self.mesh.gridFy[fyp, :]

        # Setup Mixed B.C (alpha, beta, gamma)
        temp_xm = np.ones_like(gBFxm[:, 0])
        temp_xp = np.ones_like(gBFxp[:, 0])
        temp_ym = np.ones_like(gBFym[:, 1])
        temp_yp = np.ones_like(gBFyp[:, 1])

        if self.bc_type == "Neumann":
            alpha_xm, alpha_xp = temp_xm*0., temp_xp*0.
            alpha_ym, alpha_yp = temp_ym*0., temp_yp*0.

            beta_xm, beta_xp = temp_xm, temp_xp
            beta_ym, beta_yp = temp_ym, temp_yp

            gamma_xm, gamma_xp = temp_xm*0., temp_xp*0.
            gamma_ym, gamma_yp = temp_ym*0., temp_yp*0.

        elif self.bc_type == "Dirichlet":
            alpha_xm, alpha_xp = temp_xm, temp_xp
            alpha_ym, alpha_yp = temp_ym, temp_yp

            beta_xm, beta_xp = temp_xm*0., temp_xp*0.
            beta_ym, beta_yp = temp_ym*0., temp_yp*0.

            gamma_xm, gamma_xp = temp_xm*0., temp_xp*0.
            gamma_ym, gamma_yp = temp_ym*0., temp_yp*0.

        elif self.bc_type == "Mixed":
            xs = np.median(self.mesh.vectorCCx)
            ys = np.median(self.mesh.vectorCCy[-1])

            def r_boundary(x, y):
                return 1./np.sqrt(
                    (x - xs)**2 + (y - ys)**2
                    )

            rxm = r_boundary(gBFxm[:, 0], gBFxm[:, 1])
            rxp = r_boundary(gBFxp[:, 0], gBFxp[:, 1])
            rym = r_boundary(gBFym[:, 0], gBFym[:, 1])

            alpha_xm = ky*(
                kn(1, ky*rxm) / kn(0, ky*rxm) * (gBFxm[:, 0]-xs)
                )
            alpha_xp = ky*(
                kn(1, ky*rxp) / kn(0, ky*rxp) * (gBFxp[:, 0]-xs)
                )
            alpha_ym = ky*(
                kn(1, ky*rym) / kn(0, ky*rym) * (gBFym[:, 0]-ys)
                )
            alpha_yp = temp_yp*0.
            beta_xm, beta_xp = temp_xm, temp_xp
            beta_ym, beta_yp = temp_ym, temp_yp

            gamma_xm, gamma_xp = temp_xm*0., temp_xp*0.
            gamma_ym, gamma_yp = temp_ym*0., temp_yp*0.

        alpha = [alpha_xm, alpha_xp, alpha_ym, alpha_yp]
        beta = [beta_xm, beta_xp, beta_ym, beta_yp]
        gamma = [gamma_xm, gamma_xp, gamma_ym, gamma_yp]

        x_BC, y_BC = getxBCyBC_CC(self.mesh, alpha, beta, gamma)
        V = self.Vol
        self.Div = V * self.mesh.faceDiv
        P_BC, B = self.mesh.getBCProjWF_simple()
        M = B*self.mesh.aveCC2F
        self.Grad = self.Div.T - P_BC*sdiag(y_BC)*M


class Simulation2DNodal(BaseDCSimulation2D):
    """
    2.5D nodal DC problem
    """

    _solutionType = 'phiSolution'
    _formulation = 'EB'  # CC potentials means J is on faces
    fieldsPair = Fields2DNodal
    fieldsPair_fwd = Fields3DNodal
    _gradT = None

    def __init__(self, mesh, **kwargs):
        BaseDCSimulation2D.__init__(self, mesh, **kwargs)
        # self.setBC()
        self.solver_opts['is_symmetric'] = True
        self.solver_opts['is_positive_definite'] = True

    def getA(self, ky):
        """
        Make the A matrix for the cell centered DC resistivity problem
        A = D MfRhoI G
        """
        MeSigma = self.MeSigma
        MnSigma = self.MnSigma
        Grad = self.mesh.nodalGrad
        if self._gradT is None:
            self._gradT = Grad.T.tocsr()  # cache the .tocsr()
        GradT = self._gradT
        A = GradT * MeSigma * Grad + ky**2*MnSigma
        return A

    def getADeriv(self, ky, u, v, adjoint=False):

        Grad = self.mesh.nodalGrad

        if adjoint:
            return (
                self.MeSigmaDeriv(Grad*u.flatten(), Grad*v, adjoint=adjoint) +
                ky**2*self.MnSigmaDeriv(u.flatten(), v, adjoint=adjoint)
            )
        else:
            return (
                Grad.T*self.MeSigmaDeriv(Grad*u.flatten(), v, adjoint=adjoint) +
                ky**2*self.MnSigmaDeriv(u.flatten(), v, adjoint=adjoint)
            )
        # return (Grad.T*(self.MeSigmaDeriv(Grad*u.flatten(), v, adjoint)) +
        #         ky**2*self.MnSigmaDeriv(u.flatten())*v)

    def getRHS(self, ky):
        """
        RHS for the DC problem
        q
        """

        RHS = self.getSourceTerm(ky)
        return RHS

    def getRHSDeriv(self, ky, src, v, adjoint=False):
        """
        Derivative of the right hand side with respect to the model
        """
        # TODO: add qDeriv for RHS depending on m
        # qDeriv = src.evalDeriv(self, ky, adjoint=adjoint)
        # return qDeriv
        return Zero()


Simulation2DCellCentred = Simulation2DCellCentered  # UK and US


############
# Deprecated
############

@deprecate_class(removal_version='0.15.0')
class Problem2D_N(Simulation2DNodal):
    pass


@deprecate_class(removal_version='0.15.0')
class Problem2D_CC(Simulation2DCellCentered):
    pass
